@@ -11,10 +11,10 @@
 //! The modes provided here only provide confidentiality, but **do not**
 //! provide integrity or authentication verification of ciphertext.
 //!
-//! These algorithms are provided solely for applications requring them
-//! in order to maintain backwards compatability in legacy applications.
+//! These algorithms are provided solely for applications requiring them
+//! in order to maintain backwards compatibility in legacy applications.
 //!
-//! If you are developing new applications requring data encryption see
+//! If you are developing new applications requiring data encryption see
 //! the algorithms provided in [`aead`](crate::aead).
 //!
 //! # Examples
@@ -135,6 +135,35 @@
 //! # }
 //! ```
 //!
+//! ### AES-128 CFB 128-bit mode
+//!
+//! ```rust
+//! # use std::error::Error;
+//! #
+//! # fn main() -> Result<(), Box<dyn Error>> {
+//! use aws_lc_rs::cipher::{DecryptingKey, EncryptingKey, UnboundCipherKey, AES_128};
+//!
+//! let original_message = "This is a secret message!".as_bytes();
+//! let mut in_out_buffer = Vec::from(original_message);
+//!
+//! let key_bytes: &[u8] = &[
+//!     0xff, 0x0b, 0xe5, 0x84, 0x64, 0x0b, 0x00, 0xc8, 0x90, 0x7a, 0x4b, 0xbf, 0x82, 0x7c, 0xb6,
+//!     0xd1,
+//! ];
+//!
+//! let key = UnboundCipherKey::new(&AES_128, key_bytes)?;
+//! let mut encrypting_key = EncryptingKey::cfb128(key)?;
+//! let context = encrypting_key.encrypt(&mut in_out_buffer)?;
+//!
+//! let key = UnboundCipherKey::new(&AES_128, key_bytes)?;
+//! let mut decrypting_key = DecryptingKey::cfb128(key)?;
+//! let plaintext = decrypting_key.decrypt(&mut in_out_buffer, context)?;
+//! assert_eq!(original_message, plaintext);
+//! #
+//! # Ok(())
+//! # }
+//! ```
+//!
 //! ## Constructing a `DecryptionContext` for decryption.
 //!
 //! ```rust
@@ -201,34 +230,39 @@ pub use streaming::{BufferUpdate, StreamingDecryptingKey, StreamingEncryptingKey
 
 use crate::buffer::Buffer;
 use crate::error::Unspecified;
-use crate::fips::indicator_check;
 use crate::hkdf;
 use crate::hkdf::KeyType;
 use crate::iv::{FixedLength, IV_LEN_128_BIT};
 use crate::ptr::ConstPointer;
 use aws_lc::{
-    AES_cbc_encrypt, AES_ctr128_encrypt, EVP_aes_128_cbc, EVP_aes_128_ctr, EVP_aes_256_cbc,
-    EVP_aes_256_ctr, AES_DECRYPT, AES_ENCRYPT, AES_KEY, EVP_CIPHER,
+    EVP_aes_128_cbc, EVP_aes_128_cfb128, EVP_aes_128_ctr, EVP_aes_128_ecb, EVP_aes_192_cbc,
+    EVP_aes_192_cfb128, EVP_aes_192_ctr, EVP_aes_192_ecb, EVP_aes_256_cbc, EVP_aes_256_cfb128,
+    EVP_aes_256_ctr, EVP_aes_256_ecb, EVP_CIPHER,
 };
 use core::fmt::Debug;
-use core::mem::MaybeUninit;
 use key::SymmetricCipherKey;
-use zeroize::Zeroize;
 
 /// The number of bytes in an AES 128-bit key
-pub const AES_128_KEY_LEN: usize = 16;
+pub use crate::cipher::aes::AES_128_KEY_LEN;
+
+/// The number of bytes in an AES 192-bit key
+pub use crate::cipher::aes::AES_192_KEY_LEN;
 
 /// The number of bytes in an AES 256-bit key
-pub const AES_256_KEY_LEN: usize = 32;
+pub use crate::cipher::aes::AES_256_KEY_LEN;
 
 const MAX_CIPHER_KEY_LEN: usize = AES_256_KEY_LEN;
 
 /// The number of bytes for an AES-CBC initialization vector (IV)
-pub const AES_CBC_IV_LEN: usize = 16;
+pub use crate::cipher::aes::AES_CBC_IV_LEN;
 
 /// The number of bytes for an AES-CTR initialization vector (IV)
-pub const AES_CTR_IV_LEN: usize = 16;
-const AES_BLOCK_LEN: usize = 16;
+pub use crate::cipher::aes::AES_CTR_IV_LEN;
+
+/// The number of bytes for an AES-CFB initialization vector (IV)
+pub use crate::cipher::aes::AES_CFB_IV_LEN;
+
+use crate::cipher::aes::AES_BLOCK_LEN;
 
 const MAX_CIPHER_BLOCK_LEN: usize = AES_BLOCK_LEN;
 
@@ -241,6 +275,12 @@ pub enum OperatingMode {
 
     /// Counter (CTR) mode.
     CTR,
+
+    /// CFB 128-bit mode.
+    CFB128,
+
+    /// Electronic Code Book (ECB) mode.
+    ECB,
 }
 
 impl OperatingMode {
@@ -249,8 +289,16 @@ impl OperatingMode {
         ConstPointer::new(match (self, algorithm.id) {
             (OperatingMode::CBC, AlgorithmId::Aes128) => unsafe { EVP_aes_128_cbc() },
             (OperatingMode::CTR, AlgorithmId::Aes128) => unsafe { EVP_aes_128_ctr() },
+            (OperatingMode::CFB128, AlgorithmId::Aes128) => unsafe { EVP_aes_128_cfb128() },
+            (OperatingMode::ECB, AlgorithmId::Aes128) => unsafe { EVP_aes_128_ecb() },
+            (OperatingMode::CBC, AlgorithmId::Aes192) => unsafe { EVP_aes_192_cbc() },
+            (OperatingMode::CTR, AlgorithmId::Aes192) => unsafe { EVP_aes_192_ctr() },
+            (OperatingMode::CFB128, AlgorithmId::Aes192) => unsafe { EVP_aes_192_cfb128() },
+            (OperatingMode::ECB, AlgorithmId::Aes192) => unsafe { EVP_aes_192_ecb() },
             (OperatingMode::CBC, AlgorithmId::Aes256) => unsafe { EVP_aes_256_cbc() },
             (OperatingMode::CTR, AlgorithmId::Aes256) => unsafe { EVP_aes_256_ctr() },
+            (OperatingMode::CFB128, AlgorithmId::Aes256) => unsafe { EVP_aes_256_cfb128() },
+            (OperatingMode::ECB, AlgorithmId::Aes256) => unsafe { EVP_aes_256_ecb() },
         })
         .unwrap()
     }
@@ -263,6 +311,9 @@ macro_rules! define_cipher_context {
         pub enum $name {
             /// A 128-bit Initialization Vector.
             Iv128(FixedLength<IV_LEN_128_BIT>),
+
+            /// No Cipher Context
+            None,
         }
 
         impl<'a> TryFrom<&'a $name> for &'a [u8] {
@@ -271,6 +322,7 @@ macro_rules! define_cipher_context {
             fn try_from(value: &'a $name) -> Result<Self, Unspecified> {
                 match value {
                     $name::Iv128(iv) => Ok(iv.as_ref()),
+                    _ => Err(Unspecified),
                 }
             }
         }
@@ -279,6 +331,7 @@ macro_rules! define_cipher_context {
             fn fmt(&self, f: &mut core::fmt::Formatter<'_>) -> core::fmt::Result {
                 match self {
                     Self::Iv128(_) => write!(f, "Iv128"),
+                    Self::None => write!(f, "None"),
                 }
             }
         }
@@ -287,6 +340,7 @@ macro_rules! define_cipher_context {
             fn from(value: $other) -> Self {
                 match value {
                     $other::Iv128(iv) => $name::Iv128(iv),
+                    $other::None => $name::None,
                 }
             }
         }
@@ -305,6 +359,9 @@ pub enum AlgorithmId {
 
     /// AES 256-bit
     Aes256,
+
+    /// AES 192-bit
+    Aes192,
 }
 
 /// A cipher algorithm.
@@ -322,6 +379,13 @@ pub static AES_128: Algorithm = Algorithm {
     block_len: AES_BLOCK_LEN,
 };
 
+/// AES 192-bit cipher
+pub static AES_192: Algorithm = Algorithm {
+    id: AlgorithmId::Aes192,
+    key_len: AES_192_KEY_LEN,
+    block_len: AES_BLOCK_LEN,
+};
+
 /// AES 256-bit cipher
 pub static AES_256: Algorithm = Algorithm {
     id: AlgorithmId::Aes256,
@@ -345,30 +409,40 @@ impl Algorithm {
         mode: OperatingMode,
     ) -> Result<EncryptionContext, Unspecified> {
         match self.id {
-            AlgorithmId::Aes128 | AlgorithmId::Aes256 => match mode {
-                OperatingMode::CBC | OperatingMode::CTR => {
+            // TODO: Hopefully support CFB1, and CFB8
+            AlgorithmId::Aes128 | AlgorithmId::Aes192 | AlgorithmId::Aes256 => match mode {
+                OperatingMode::CBC | OperatingMode::CTR | OperatingMode::CFB128 => {
                     Ok(EncryptionContext::Iv128(FixedLength::new()?))
                 }
+                OperatingMode::ECB => Ok(EncryptionContext::None),
             },
         }
     }
 
     fn is_valid_encryption_context(&self, mode: OperatingMode, input: &EncryptionContext) -> bool {
         match self.id {
-            AlgorithmId::Aes128 | AlgorithmId::Aes256 => match mode {
-                OperatingMode::CBC | OperatingMode::CTR => {
+            // TODO: Hopefully support CFB1, and CFB8
+            AlgorithmId::Aes128 | AlgorithmId::Aes192 | AlgorithmId::Aes256 => match mode {
+                OperatingMode::CBC | OperatingMode::CTR | OperatingMode::CFB128 => {
                     matches!(input, EncryptionContext::Iv128(_))
                 }
+                OperatingMode::ECB => {
+                    matches!(input, EncryptionContext::None)
+                }
             },
         }
     }
 
     fn is_valid_decryption_context(&self, mode: OperatingMode, input: &DecryptionContext) -> bool {
+        // TODO: Hopefully support CFB1, and CFB8
         match self.id {
-            AlgorithmId::Aes128 | AlgorithmId::Aes256 => match mode {
-                OperatingMode::CBC | OperatingMode::CTR => {
+            AlgorithmId::Aes128 | AlgorithmId::Aes192 | AlgorithmId::Aes256 => match mode {
+                OperatingMode::CBC | OperatingMode::CTR | OperatingMode::CFB128 => {
                     matches!(input, DecryptionContext::Iv128(_))
                 }
+                OperatingMode::ECB => {
+                    matches!(input, DecryptionContext::None)
+                }
             },
         }
     }
@@ -433,6 +507,7 @@ impl TryInto<SymmetricCipherKey> for UnboundCipherKey {
     fn try_into(self) -> Result<SymmetricCipherKey, Self::Error> {
         match self.algorithm.id() {
             AlgorithmId::Aes128 => SymmetricCipherKey::aes128(self.key_bytes.as_ref()),
+            AlgorithmId::Aes192 => SymmetricCipherKey::aes192(self.key_bytes.as_ref()),
             AlgorithmId::Aes256 => SymmetricCipherKey::aes256(self.key_bytes.as_ref()),
         }
     }
@@ -455,15 +530,45 @@ impl EncryptingKey {
     //
     /// # Errors
     /// * [`Unspecified`]: Returned if there is an error constructing the `EncryptingKey`.
-    pub fn ctr(key: UnboundCipherKey) -> Result<EncryptingKey, Unspecified> {
-        EncryptingKey::new(key, OperatingMode::CTR)
+    pub fn ctr(key: UnboundCipherKey) -> Result<Self, Unspecified> {
+        Self::new(key, OperatingMode::CTR)
+    }
+
+    /// Constructs an `EncryptingKey` operating in cipher feedback 128-bit mode (CFB128) using the provided key.
+    ///
+    // # FIPS
+    // Use this function with an `UnboundCipherKey` constructed with one of the following algorithms:
+    // * `AES_128`
+    // * `AES_256`
+    //
+    /// # Errors
+    /// * [`Unspecified`]: Returned if there is an error constructing the `EncryptingKey`.
+    pub fn cfb128(key: UnboundCipherKey) -> Result<Self, Unspecified> {
+        Self::new(key, OperatingMode::CFB128)
+    }
+
+    /// Constructs an `EncryptingKey` operating in electronic code book mode (ECB) using the provided key.
+    ///
+    /// # ☠️ ️️️DANGER ☠️
+    /// Offered for computability purposes only. This is an extremely dangerous mode, and
+    /// very likely not what you want to use.
+    ///
+    // # FIPS
+    // Use this function with an `UnboundCipherKey` constructed with one of the following algorithms:
+    // * `AES_128`
+    // * `AES_256`
+    //
+    /// # Errors
+    /// * [`Unspecified`]: Returned if there is an error constructing the `EncryptingKey`.
+    pub fn ecb(key: UnboundCipherKey) -> Result<Self, Unspecified> {
+        Self::new(key, OperatingMode::ECB)
     }
 
     #[allow(clippy::unnecessary_wraps)]
-    fn new(key: UnboundCipherKey, mode: OperatingMode) -> Result<EncryptingKey, Unspecified> {
+    fn new(key: UnboundCipherKey, mode: OperatingMode) -> Result<Self, Unspecified> {
         let algorithm = key.algorithm();
         let key = key.try_into()?;
-        Ok(EncryptingKey {
+        Ok(Self {
             algorithm,
             key,
             mode,
@@ -486,6 +591,9 @@ impl EncryptingKey {
     /// Returns a [`DecryptionContext`] with the randomly generated IV that was used to encrypt
     /// the data provided.
     ///
+    /// If `EncryptingKey` is operating in `OperatingMode::ECB`, then `in_out.len()` must be a multiple
+    /// of the block length.
+    ///
     /// # Errors
     /// * [`Unspecified`]: Returned if cipher mode requires input to be a multiple of the block length,
     ///   and `in_out.len()` is not. Otherwise, returned if encryption fails.
@@ -499,6 +607,9 @@ impl EncryptingKey {
     /// a `EncryptionContext` from a previously used IV (initialization vector).
     /// Returns a [`DecryptionContext`] produced from the provided `EncryptionContext`.
     ///
+    /// If `EncryptingKey` is operating in `OperatingMode::ECB`, then `in_out.len()` must be a multiple
+    /// of the block length.
+    ///
     /// # Errors
     /// * [`Unspecified`]: Returned if cipher mode requires input to be a multiple of the block length,
     ///   and `in_out.len()` is not. Otherwise returned if encryption fails.
@@ -544,14 +655,44 @@ impl DecryptingKey {
     /// # Errors
     /// * [`Unspecified`]: Returned if there is an error during decryption.
     pub fn ctr(key: UnboundCipherKey) -> Result<DecryptingKey, Unspecified> {
-        DecryptingKey::new(key, OperatingMode::CTR)
+        Self::new(key, OperatingMode::CTR)
+    }
+
+    /// Constructs a cipher decrypting key operating in cipher feedback 128-bit mode (CFB128) using the provided key and context.
+    ///
+    // # FIPS
+    // Use this function with an `UnboundCipherKey` constructed with one of the following algorithms:
+    // * `AES_128`
+    // * `AES_256`
+    //
+    /// # Errors
+    /// * [`Unspecified`]: Returned if there is an error during decryption.
+    pub fn cfb128(key: UnboundCipherKey) -> Result<Self, Unspecified> {
+        Self::new(key, OperatingMode::CFB128)
+    }
+
+    /// Constructs an `DecryptingKey` operating in electronic code book (ECB) mode using the provided key.
+    ///
+    /// # ☠️ ️️️DANGER ☠️
+    /// Offered for computability purposes only. This is an extremely dangerous mode, and
+    /// very likely not what you want to use.
+    ///
+    // # FIPS
+    // Use this function with an `UnboundCipherKey` constructed with one of the following algorithms:
+    // * `AES_128`
+    // * `AES_256`
+    //
+    /// # Errors
+    /// * [`Unspecified`]: Returned if there is an error constructing the `DecryptingKey`.
+    pub fn ecb(key: UnboundCipherKey) -> Result<Self, Unspecified> {
+        Self::new(key, OperatingMode::ECB)
     }
 
     #[allow(clippy::unnecessary_wraps)]
-    fn new(key: UnboundCipherKey, mode: OperatingMode) -> Result<DecryptingKey, Unspecified> {
+    fn new(key: UnboundCipherKey, mode: OperatingMode) -> Result<Self, Unspecified> {
         let algorithm = key.algorithm();
         let key = key.try_into()?;
-        Ok(DecryptingKey {
+        Ok(Self {
             algorithm,
             key,
             mode,
@@ -573,6 +714,9 @@ impl DecryptingKey {
     /// Decrypts the data provided in `in_out` in-place.
     /// Returns a references to the decrypted data.
     ///
+    /// If `DecryptingKey` is operating in `OperatingMode::ECB`, then `in_out.len()` must be a multiple
+    /// of the block length.
+    ///
     /// # Errors
     /// * [`Unspecified`]: Returned if cipher mode requires input to be a multiple of the block length,
     ///   and `in_out.len()` is not. Also returned if decryption fails.
@@ -604,20 +748,35 @@ fn encrypt(
     let block_len = algorithm.block_len();
 
     match mode {
-        OperatingMode::CTR => {}
-        _ => {
-            if (in_out.len() % block_len) != 0 {
+        OperatingMode::CBC | OperatingMode::ECB => {
+            if in_out.len() % block_len != 0 {
                 return Err(Unspecified);
             }
         }
+        _ => {}
     }
 
     match mode {
         OperatingMode::CBC => match algorithm.id() {
-            AlgorithmId::Aes128 | AlgorithmId::Aes256 => encrypt_aes_cbc_mode(key, context, in_out),
+            AlgorithmId::Aes128 | AlgorithmId::Aes192 | AlgorithmId::Aes256 => {
+                aes::encrypt_cbc_mode(key, context, in_out)
+            }
         },
         OperatingMode::CTR => match algorithm.id() {
-            AlgorithmId::Aes128 | AlgorithmId::Aes256 => encrypt_aes_ctr_mode(key, context, in_out),
+            AlgorithmId::Aes128 | AlgorithmId::Aes192 | AlgorithmId::Aes256 => {
+                aes::encrypt_ctr_mode(key, context, in_out)
+            }
+        },
+        // TODO: Hopefully support CFB1, and CFB8
+        OperatingMode::CFB128 => match algorithm.id() {
+            AlgorithmId::Aes128 | AlgorithmId::Aes192 | AlgorithmId::Aes256 => {
+                aes::encrypt_cfb_mode(key, mode, context, in_out)
+            }
+        },
+        OperatingMode::ECB => match algorithm.id() {
+            AlgorithmId::Aes128 | AlgorithmId::Aes192 | AlgorithmId::Aes256 => {
+                aes::encrypt_ecb_mode(key, context, in_out)
+            }
         },
     }
 }
@@ -632,155 +791,39 @@ fn decrypt<'in_out>(
     let block_len = algorithm.block_len();
 
     match mode {
-        OperatingMode::CTR => {}
-        _ => {
-            if (in_out.len() % block_len) != 0 {
+        OperatingMode::CBC | OperatingMode::ECB => {
+            if in_out.len() % block_len != 0 {
                 return Err(Unspecified);
             }
         }
+        _ => {}
     }
 
     match mode {
         OperatingMode::CBC => match algorithm.id() {
-            AlgorithmId::Aes128 | AlgorithmId::Aes256 => decrypt_aes_cbc_mode(key, context, in_out),
+            AlgorithmId::Aes128 | AlgorithmId::Aes192 | AlgorithmId::Aes256 => {
+                aes::decrypt_cbc_mode(key, context, in_out)
+            }
         },
         OperatingMode::CTR => match algorithm.id() {
-            AlgorithmId::Aes128 | AlgorithmId::Aes256 => decrypt_aes_ctr_mode(key, context, in_out),
+            AlgorithmId::Aes128 | AlgorithmId::Aes192 | AlgorithmId::Aes256 => {
+                aes::decrypt_ctr_mode(key, context, in_out)
+            }
+        },
+        // TODO: Hopefully support CFB1, and CFB8
+        OperatingMode::CFB128 => match algorithm.id() {
+            AlgorithmId::Aes128 | AlgorithmId::Aes192 | AlgorithmId::Aes256 => {
+                aes::decrypt_cfb_mode(key, mode, context, in_out)
+            }
+        },
+        OperatingMode::ECB => match algorithm.id() {
+            AlgorithmId::Aes128 | AlgorithmId::Aes192 | AlgorithmId::Aes256 => {
+                aes::decrypt_ecb_mode(key, context, in_out)
+            }
         },
     }
 }
 
-fn encrypt_aes_ctr_mode(
-    key: &SymmetricCipherKey,
-    context: EncryptionContext,
-    in_out: &mut [u8],
-) -> Result<DecryptionContext, Unspecified> {
-    #[allow(clippy::match_wildcard_for_single_variants)]
-    let key = match &key {
-        SymmetricCipherKey::Aes128 { enc_key, .. } | SymmetricCipherKey::Aes256 { enc_key, .. } => {
-            enc_key
-        }
-        _ => return Err(Unspecified),
-    };
-
-    let mut iv = {
-        let mut iv = [0u8; AES_CTR_IV_LEN];
-        iv.copy_from_slice((&context).try_into()?);
-        iv
-    };
-
-    let mut buffer = [0u8; AES_BLOCK_LEN];
-
-    aes_ctr128_encrypt(key, &mut iv, &mut buffer, in_out);
-    iv.zeroize();
-
-    Ok(context.into())
-}
-
-fn decrypt_aes_ctr_mode<'in_out>(
-    key: &SymmetricCipherKey,
-    context: DecryptionContext,
-    in_out: &'in_out mut [u8],
-) -> Result<&'in_out mut [u8], Unspecified> {
-    // it's the same in CTR, just providing a nice named wrapper to match
-    encrypt_aes_ctr_mode(key, context.into(), in_out).map(|_| in_out)
-}
-
-fn encrypt_aes_cbc_mode(
-    key: &SymmetricCipherKey,
-    context: EncryptionContext,
-    in_out: &mut [u8],
-) -> Result<DecryptionContext, Unspecified> {
-    #[allow(clippy::match_wildcard_for_single_variants)]
-    let key = match &key {
-        SymmetricCipherKey::Aes128 { enc_key, .. } | SymmetricCipherKey::Aes256 { enc_key, .. } => {
-            enc_key
-        }
-        _ => return Err(Unspecified),
-    };
-
-    let mut iv = {
-        let mut iv = [0u8; AES_CBC_IV_LEN];
-        iv.copy_from_slice((&context).try_into()?);
-        iv
-    };
-
-    aes_cbc_encrypt(key, &mut iv, in_out);
-    iv.zeroize();
-
-    Ok(context.into())
-}
-
-#[allow(clippy::needless_pass_by_value)]
-fn decrypt_aes_cbc_mode<'in_out>(
-    key: &SymmetricCipherKey,
-    context: DecryptionContext,
-    in_out: &'in_out mut [u8],
-) -> Result<&'in_out mut [u8], Unspecified> {
-    #[allow(clippy::match_wildcard_for_single_variants)]
-    let key = match &key {
-        SymmetricCipherKey::Aes128 { dec_key, .. } | SymmetricCipherKey::Aes256 { dec_key, .. } => {
-            dec_key
-        }
-        _ => return Err(Unspecified),
-    };
-
-    let mut iv = {
-        let mut iv = [0u8; AES_CBC_IV_LEN];
-        iv.copy_from_slice((&context).try_into()?);
-        iv
-    };
-
-    aes_cbc_decrypt(key, &mut iv, in_out);
-    iv.zeroize();
-
-    Ok(in_out)
-}
-
-fn aes_ctr128_encrypt(key: &AES_KEY, iv: &mut [u8], block_buffer: &mut [u8], in_out: &mut [u8]) {
-    let mut num = MaybeUninit::<u32>::new(0);
-
-    indicator_check!(unsafe {
-        AES_ctr128_encrypt(
-            in_out.as_ptr(),
-            in_out.as_mut_ptr(),
-            in_out.len(),
-            key,
-            iv.as_mut_ptr(),
-            block_buffer.as_mut_ptr(),
-            num.as_mut_ptr(),
-        );
-    });
-
-    Zeroize::zeroize(block_buffer);
-}
-
-fn aes_cbc_encrypt(key: &AES_KEY, iv: &mut [u8], in_out: &mut [u8]) {
-    indicator_check!(unsafe {
-        AES_cbc_encrypt(
-            in_out.as_ptr(),
-            in_out.as_mut_ptr(),
-            in_out.len(),
-            key,
-            iv.as_mut_ptr(),
-            AES_ENCRYPT,
-        );
-    });
-}
-
-fn aes_cbc_decrypt(key: &AES_KEY, iv: &mut [u8], in_out: &mut [u8]) {
-    indicator_check!(unsafe {
-        AES_cbc_encrypt(
-            in_out.as_ptr(),
-            in_out.as_mut_ptr(),
-            in_out.len(),
-            key,
-            iv.as_mut_ptr(),
-            AES_DECRYPT,
-        );
-    });
-}
-
 #[cfg(test)]
 mod tests {
     use super::*;
@@ -874,6 +917,23 @@ mod tests {
         }
     }
 
+    #[test]
+    fn test_aes_128_cfb128() {
+        let key = from_hex("000102030405060708090a0b0c0d0e0f").unwrap();
+        for i in 0..=50 {
+            helper_test_cipher_n_bytes(key.as_slice(), &AES_128, OperatingMode::CFB128, i);
+        }
+    }
+
+    #[test]
+    fn test_aes_256_cfb128() {
+        let key =
+            from_hex("000102030405060708090a0b0c0d0e0f000102030405060708090a0b0c0d0e0f").unwrap();
+        for i in 0..=50 {
+            helper_test_cipher_n_bytes(key.as_slice(), &AES_256, OperatingMode::CFB128, i);
+        }
+    }
+
     #[test]
     fn test_aes_256_ctr() {
         let key =
@@ -883,6 +943,12 @@ mod tests {
         }
     }
 
+    #[test]
+    fn test_aes_128_ecb() {
+        let key = from_hex("000102030405060708090a0b0c0d0e0f").unwrap();
+        _ = key;
+    }
+
     macro_rules! cipher_kat {
         ($name:ident, $alg:expr, $mode:expr, $key:literal, $iv: literal, $plaintext:literal, $ciphertext:literal) => {
             #[test]
@@ -918,6 +984,34 @@ mod tests {
                 let unbound_key2 = UnboundCipherKey::new(alg, &key).unwrap();
                 let decrypting_key = DecryptingKey::new(unbound_key2, $mode).unwrap();
 
+                let plaintext = decrypting_key.decrypt(&mut in_out, context).unwrap();
+                assert_eq!(input.as_slice(), plaintext);
+            }
+        };
+        ($name:ident, $alg:expr, $mode:expr, $key:literal, $plaintext:literal, $ciphertext:literal) => {
+            #[test]
+            fn $name() {
+                let key = from_hex($key).unwrap();
+                let input = from_hex($plaintext).unwrap();
+                let expected_ciphertext = from_hex($ciphertext).unwrap();
+
+                let alg = $alg;
+
+                let unbound_key = UnboundCipherKey::new(alg, &key).unwrap();
+
+                let encrypting_key = EncryptingKey::new(unbound_key, $mode).unwrap();
+
+                let mut in_out = input.clone();
+
+                let context = encrypting_key
+                    .less_safe_encrypt(&mut in_out, EncryptionContext::None)
+                    .unwrap();
+
+                assert_eq!(expected_ciphertext, in_out);
+
+                let unbound_key2 = UnboundCipherKey::new(alg, &key).unwrap();
+                let decrypting_key = DecryptingKey::new(unbound_key2, $mode).unwrap();
+
                 let plaintext = decrypting_key.decrypt(&mut in_out, context).unwrap();
                 assert_eq!(input.as_slice(), plaintext);
             }
@@ -963,4 +1057,42 @@ mod tests {
         "eca7285d19f3c20e295378460e8729",
         "b5098e5e788de6ac2f2098eb2fc6f8"
     );
+
+    cipher_kat!(
+        test_sp800_38a_cfb128_aes128,
+        &AES_128,
+        OperatingMode::CFB128,
+        "2b7e151628aed2a6abf7158809cf4f3c",
+        "000102030405060708090a0b0c0d0e0f",
+        "6bc1bee22e409f96e93d7e117393172aae2d8a571e03ac9c9eb76fac45af8e5130c81c46a35ce411e5fbc1191a0a52eff69f2445df4f9b17ad2b417be66c3710",
+        "3b3fd92eb72dad20333449f8e83cfb4ac8a64537a0b3a93fcde3cdad9f1ce58b26751f67a3cbb140b1808cf187a4f4dfc04b05357c5d1c0eeac4c66f9ff7f2e6"
+    );
+
+    cipher_kat!(
+        test_sp800_38a_cfb128_aes256,
+        &AES_256,
+        OperatingMode::CFB128,
+        "603deb1015ca71be2b73aef0857d77811f352c073b6108d72d9810a30914dff4",
+        "000102030405060708090a0b0c0d0e0f",
+        "6bc1bee22e409f96e93d7e117393172aae2d8a571e03ac9c9eb76fac45af8e5130c81c46a35ce411e5fbc1191a0a52eff69f2445df4f9b17ad2b417be66c3710",
+        "dc7e84bfda79164b7ecd8486985d386039ffed143b28b1c832113c6331e5407bdf10132415e54b92a13ed0a8267ae2f975a385741ab9cef82031623d55b1e471"
+    );
+
+    cipher_kat!(
+        test_sp800_38a_ecb_aes128,
+        &AES_128,
+        OperatingMode::ECB,
+        "2b7e151628aed2a6abf7158809cf4f3c",
+        "6bc1bee22e409f96e93d7e117393172aae2d8a571e03ac9c9eb76fac45af8e5130c81c46a35ce411e5fbc1191a0a52eff69f2445df4f9b17ad2b417be66c3710",
+        "3ad77bb40d7a3660a89ecaf32466ef97f5d3d58503b9699de785895a96fdbaaf43b1cd7f598ece23881b00e3ed0306887b0c785e27e8ad3f8223207104725dd4"
+    );
+
+    cipher_kat!(
+        test_sp800_38a_ecb_aes256,
+        &AES_256,
+        OperatingMode::ECB,
+        "603deb1015ca71be2b73aef0857d77811f352c073b6108d72d9810a30914dff4",
+        "6bc1bee22e409f96e93d7e117393172aae2d8a571e03ac9c9eb76fac45af8e5130c81c46a35ce411e5fbc1191a0a52eff69f2445df4f9b17ad2b417be66c3710",
+        "f3eed1bdb5d2a03c064b5a7e3db181f8591ccb10d410ed26dc5ba74a31362870b6ed21b99ca6f4f9f153e7b1beafed1d23304b7a39f9f3ff067d8d8f9e24ecc7"
+    );
 }
